package fsl;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import mgui.command.CommandFunctions;
import mgui.command.CommandFunctions.ProcessState;
import mgui.command.CommandInstance;
import mgui.interfaces.logs.LoggingType;
import mgui.interfaces.maps.NameMap;
import mgui.io.domestic.maps.NameMapLoader;
import mgui.io.util.IoFunctions;
import mgui.util.StringFunctions;

/******************************************************
 * Utility class to perform FSL functions
 * 
 * @author Ardrew Reid
 *
 */
public class FslFunctions extends CommandInstance {

		public static String path = null;
		
		public static void main(String[] args) throws Exception{
			
			FslFunctions instance = new FslFunctions();
			instance.execute(args);
			
		}

		
		@Override
		protected boolean run_command(String command) {
			
			if (command.equals("generate_probtrackx_matrices")){
				return generate_probtrackx_matrices();
				}
			
			
			
			log("FslFunctions: no such command '" + command + "'.",
					LoggingType.Errors);
			return false;
		}

		/*******************************************************
		 * Takes a set of probtrackx parameters, plus the ones below, and generates a matrix file for every
		 * ROI seed in a name map file. Use {subject} and {roi} to insert these values in the probtrackx parameters.
		 * 
		 * Parameters:
		 * 
		 * subject_dir: 				Directory containing the subject data in subdirectories
		 * name_map_file: 				Name map file from which to derive the ROI names
		 * matrix_file_pattern: 		Pattern defining the name of the output matrix files; use {subject} to
		 * 								insert the subject id, {roi} to insert the ROI name
		 * environment_file: 			[optional] File specifying the FSL environment (String pairs)
		 * path: 						Path to FSL dir
		 * skip_existing: 				Skip existing subjects
		 * 
		 * @return
		 */
		boolean generate_probtrackx_matrices(){
			
			// Get probtrackx parameters spearately
			HashMap<String,String> probtrackx_params = new HashMap<String,String>(parameters);
			probtrackx_params.remove("matrix_file_pattern");
			probtrackx_params.remove("environment_file");
			probtrackx_params.remove("path");
			probtrackx_params.remove("name_map_file");
			probtrackx_params.remove("subject_dir");
			probtrackx_params.remove("skip_existing");
			
			boolean skip_existing = parameters.containsKey("skip_existing");
			
			String fsl_dir = parameters.get("path");
			if (fsl_dir == null){
				log("generate_probtrackx_matrices: parameter path not optional.",
						LoggingType.Errors);
					return false;
				}
			
			if (fsl_dir.endsWith(File.separator))
				fsl_dir = fsl_dir.substring(0, fsl_dir.length() - 1);
			
			String _param = parameters.get("environment_file");
			
			HashMap<String,String> environment = null;
			
			if (_param != null){
				try{
					environment = loadFslEnvironment(_param, fsl_dir);
				}catch (IOException ex){
					log("generate_probtrackx_matrices: Could not load environment: " + ex.getMessage(),
							LoggingType.Errors);
						return false;
					}
				}
			
			_param = parameters.get("subject_dir");
			
			if (_param == null){
				log("generate_probtrackx_matrices: parameter subject_dir not optional.",
						LoggingType.Errors);
					return false;
				}
			File subject_dir = new File(_param);
			if (!subject_dir.exists()){
				log("generate_probtrackx_matrices: subject_dir '" + subject_dir.getAbsolutePath() + "' doesn't exist.",
					LoggingType.Errors);
				return false;
				} 
			
			_param = parameters.get("name_map_file");
			
			if (_param == null){
				log("generate_probtrackx_matrices: parameter name_map_file not optional.",
						LoggingType.Errors);
					return false;
				}
			File name_map_file = new File(_param);
			if (!name_map_file.exists()){
				log("generate_probtrackx_matrices: name_map_file '" + name_map_file.getAbsolutePath() + "' doesn't exist.",
					LoggingType.Errors);
				return false;
				}
			
			NameMapLoader nm_loader = new NameMapLoader(name_map_file);
			NameMap name_map = null;
			
			try{
				name_map = (NameMap)nm_loader.loadObject();
			}catch (IOException ex){
				log("generate_probtrackx_matrices: Erroring reading name map file '" + name_map_file.getAbsolutePath() + "'.\nDetails: " +
					ex.getMessage(),
					LoggingType.Errors);
					return false;
				}
			
			ArrayList<String> roi_names = new ArrayList<String>(name_map.getNames());
			Collections.sort(roi_names);
			
			String output_dir_pattern = parameters.get("dir");
			if (output_dir_pattern == null){
				log("generate_probtrackx_matrices: parameter dir not optional.",
						LoggingType.Errors);
					return false;
				}
			
			String samples_file_pattern = parameters.get("samples");
			if (samples_file_pattern == null){
				log("generate_probtrackx_matrices: parameter samples not optional.",
						LoggingType.Errors);
					return false;
				}
			
			String mask_file_pattern = parameters.get("mask");
			if (mask_file_pattern == null){
				log("generate_probtrackx_matrices: parameter mask not optional.",
						LoggingType.Errors);
					return false;
				}
			
			String matrix_file_pattern = parameters.get("matrix_file_pattern");
			if (matrix_file_pattern == null){
				log("generate_probtrackx_matrices: parameter matrix_file_pattern not optional.",
						LoggingType.Errors);
					return false;
				}
			
			String roi_mask_file_pattern = parameters.get("seed");
			if (roi_mask_file_pattern == null){
				log("generate_probtrackx_matrices: parameter seed not optional.",
						LoggingType.Errors);
					return false;
				}
			
			
			int success_count = 0, fail_count = 0, skip_count = 0;
			
			String[] subjects = subject_dir.list(IoFunctions.getDirFilter());
			
			// For all subjects
			for (int i = 0; i < subjects.length; i++){
				
				if (this.is_interrupted){
					log("generate_probtrackx_matrices: Process was interrupted.",
							LoggingType.Errors);
					return false;
					}
				
				boolean failed = false;
				
				String subject = subjects[i];
				String subject_subdir = subject_dir.getAbsolutePath() + File.separator +  subject;
				// Insert subject id into file patterns
				String s_samples_file = StringFunctions.replace(samples_file_pattern, "{subject}", subject);
				String s_mask_file = StringFunctions.replace(mask_file_pattern, "{subject}", subject);
				String s_output_dir = StringFunctions.replace(output_dir_pattern, "{subject}", subject);
								
				String s_roi_mask_file_pattern = StringFunctions.replace(roi_mask_file_pattern, "{subject}", subject);
				
				// For all ROIS, run probtrackx:
				for (int j = 0; j < roi_names.size() && !failed; j++){
					
					if (this.is_interrupted){
						log("generate_probtrackx_matrices: Process was interrupted.",
								LoggingType.Errors);
						return false;
						}
					
					boolean skip = false;
					
					String roi = roi_names.get(j);
					String roi_output_dir = s_output_dir + File.separator + roi;
					if ((new File(roi_output_dir)).exists() && skip_existing)
						skip = true;
					// 1. Update probtrackx values with roi
					String seed_roi = subject_subdir + File.separator + StringFunctions.replace(s_roi_mask_file_pattern, "{roi}", roi);
					
					// 2. Generate list of other mask images
					File targets_file = null;
					if (!failed && !skip){
						try{
							targets_file = File.createTempFile("seed_list", ".txt");
							BufferedWriter writer = new BufferedWriter(new FileWriter(targets_file));
							for (int k = 0; k < roi_names.size(); k++){
								if (k != j){
									String this_roi = roi_names.get(k);
									String file_name = subject_subdir + File.separator + StringFunctions.replace(s_roi_mask_file_pattern, "{roi}", this_roi);
									writer.write(file_name + "\n");
									}
								}
							writer.close();
						}catch (IOException ex){
							log("generate_probtrackx_matrices: failed to write to temp file '" + targets_file.getAbsolutePath() +"'.",
									LoggingType.Errors);
								failed = true;
							}
						}
					
					if (!failed && ! skip){
						// 3. Set parameter values
						probtrackx_params.put("seed", seed_roi);
						probtrackx_params.put("dir", roi_output_dir);
						probtrackx_params.put("targetmasks", targets_file.getAbsolutePath());
						probtrackx_params.put("samples", s_samples_file);
						probtrackx_params.put("mask", s_mask_file);
						
						// 4. Run process and wait
						String command = fsl_dir + File.separator + "bin" + File.separator + "probtrackx " +  getFslParameters(probtrackx_params);
						
						ProcessState state = ProcessState.NotStarted;
						
						try{
							state = CommandFunctions.callAndWait("probtrackx", command, -1, environment);
						}catch (IOException ex){
							state = ProcessState.TerminatedWithError;
							log("generate_probtrackx_matrices: Exception while running command: " + ex.getMessage(),
									LoggingType.Errors);
							failed = true;
							}
						
						if (state != ProcessState.TerminatedNormally){
							failed = true;
							}
						}
					if (!failed){
						log(roi + ": success.");
					}else if (skip){
						log(roi + ": skipped existing.");
					}else{
						log(roi + ": failed.", LoggingType.Errors);
						}
					}
				
				if (!failed){
					success_count++;
					log("Subject " + subject + ": success.");
				}else{
					fail_count++;
					log("Subject " + subject + ": failed.", LoggingType.Errors);
					}
				
				
				}
			
			return true;
			
		}
		
		String getFslParameters(HashMap<String,String> params){
			
			ArrayList<String> keys = new ArrayList<String>(params.keySet());
			String p_list = "";
			
			for (int i = 0; i < keys.size(); i++){
				String key = keys.get(i);
				String value = params.get(key);
				if (i > 0) p_list = p_list + " ";
				p_list = p_list + "--" + key;
				if (value != null && value.length() > 0)
					p_list = p_list + "=" + value;
				}
			
			return p_list;
		}
		
		HashMap<String,String> loadFslEnvironment(String file, String fsl_dir) throws IOException{
			
			HashMap<String,String> environment = new HashMap<String,String>();
			
			File e_file = new File(file);
			BufferedReader reader = new BufferedReader(new FileReader(e_file));
			
			String line = reader.readLine();
			
			while (line != null){
				if (!line.contains(" "))
					throw new IOException("Environment file must be a list of space-delimited values..");
				line = StringFunctions.replace(line, "$FSLDIR", fsl_dir);
				String[] vals = line.split(" ");
				if (vals.length == 2)
					environment.put(vals[0], vals[1]);
				else
					environment.put(vals[0], "");
				line = reader.readLine();
				}
			
			reader.close();
			return environment;
			
		}


		
}
