/*
* Copyright (C) 2011 Andrew Reid and the modelGUI Project <http://mgui.wikidot.com>
* 
* This file is part of modelGUI[neuro] (mgui-neuro).
* 
* modelGUI[core] is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* modelGUI[core] is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with modelGUI[neuro]. If not, see <http://www.gnu.org/licenses/>.
*/

package mgui.command;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.StringTokenizer;
import java.util.TreeSet;

import mgui.geometry.Mesh3D;
import mgui.geometry.mesh.MeshFunctions;
import mgui.interfaces.logs.LoggingType;
import mgui.interfaces.maps.NameMap;
import mgui.interfaces.shapes.Mesh3DInt;
import mgui.io.domestic.maps.NameMapLoader;
import mgui.io.foreign.minc.MincSurfaceDataLoader;
import mgui.io.foreign.minc.MincSurfaceLoader;
import mgui.io.util.IoFunctions;
import mgui.numbers.MguiDouble;
import mgui.numbers.MguiNumber;
import mgui.util.Colours;

/*************************************************
 * Provides a set of functions for performing specific tasks on Civet output. These processes form the
 * <strong>mgui-civet</strong> process library.
 * 
 * @author Andrew Reid
 * @version 1.0
 * @since 1.0
 *
 */
public class CivetFunctions extends CommandInstance {

	public static void main(String[] args) throws Exception{
		
		CivetFunctions instance = new CivetFunctions();
		instance.execute(args);
		
	}
	
	@Override
	public boolean run_command(String command){
		
		if (command.equals("average-surface-data")){
			return average_surface_data();
			}
		
		if (command.equals("stdev-surface-data")){
			return stdev_surface_data();
			}
		
		if (command.equals("calculate-surface-area")){
			return calculate_area_and_volume();
			}
		
		if (command.equals("vertex-scatterplot")){
			return vertex_scatterplot();
			}
		
		if (command.equals("roi-scatterplots")){
			return roi_scatterplots();
			}
		
		if (command.equals("convert-integer-values")){
			return convert_integer_values();
			}
		
		return false;
	}
	
	/*********************************************
	 * Calculates a vertex-wise average for a set of surface data instances (e.g., cortical thickness).
	 * Instances are specified by an instance directory and a "\n"-delimited text file listing instance 
	 * names. If no list file is specified, all instances are used. 
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>instance_dir			The instance directory
	 * <li>input_file_pattern	Pattern specifying a path to each input data file, relative to each instance 
	 * 							subdirectory. Use "{subject}" to specify insert the current instance name into 
	 * 							the path.
	 * <li>output_file			The file to contain the averaged vertex values
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>instance_list_file	A text file listing the instances to average ("\n"-delimited); no file
	 * 							indicates that all instances are to be used
	 * <li>log_file				File to place log comments and errors; no log_file specifies that comments
	 * 							are to be sent to the console
	 * <li>number_format		Number format, e.g., "##0.000".
	 * </ul>
	 * 
	 */
	public boolean average_surface_data(){
		
		// Get subdirectories
		String param = parameters.get("instance_dir");
		if (param == null){
			log("average_surface_data: Parameter 'instance_dir' is required.");
			return false;
			}
		
		File instance_dir = new File(param);
		
		if (!instance_dir.exists() || !instance_dir.isDirectory()){
			log("average_surface_data: Directory '" + instance_dir.getAbsolutePath() + "' does not " +
														"exist or is not a directory.");
			return false;
			}
		
		String input_pattern = parameters.get("input_file_pattern");
		if (input_pattern == null){
			log("average_surface_data: Parameter 'input_file_pattern' is required.");
			return false;
			}
		
		input_pattern = input_pattern.replace("/", File.separator);
		input_pattern = input_pattern.replace("\\", File.separator);
		input_pattern = param + File.separator + input_pattern;
		
		param = parameters.get("output_file");
		if (param == null){
			log("average_surface_data: Parameter 'output_file' is required.");
			return false;
			}
		
		File output_file = new File(param);
		
		try{
			if (!output_file.exists() && !output_file.createNewFile()){
				log("average_surface_data: File '" + output_file.getAbsolutePath() + "' could not be created.");
				return false;
				}
		}catch (IOException e){
			log("average_surface_data: File '" + output_file.getAbsolutePath() + "' could not be created.");
			return false;
			}
		
		String number_format = parameters.get("number_format");
		if (number_format == null){
			number_format = "#0.00000000";
			}
		
		ArrayList<String> instances = new ArrayList<String>();
		
		// Load subject list
		param = parameters.get("instance_list_file");
		if (param == null){
			log("average_surface_data: No instance file specified; using all instances.");
			String[] list = instance_dir.list(IoFunctions.getDirFilter());
			for (int i = 0; i < list.length; i++)
				instances.add(list[i]);
		}else{
			File list_file = new File(param);
			log("average_surface_data: Loading instances from '" + list_file.getAbsolutePath() + "'.");
			if (!list_file.exists()){
				log("average_surface_data: File '" + list_file.getAbsolutePath() + "' not found.");
				return false;
				}
			try{
				BufferedReader reader = new BufferedReader(new FileReader(list_file));
				String line = reader.readLine();
				while (line != null){
					instances.add(line);
					line = reader.readLine();
					}
				reader.close();
					
			}catch (IOException e){
				log("average_surface_data: Error reading file '" + list_file.getAbsolutePath() + "'.");
				return false;
				}
			}
		
		// Cross-reference; report errors
		
		
		// For each instance, load data, add to sums
		ArrayList<Double> sums = new ArrayList<Double>();
		int success_count = 0, fail_count = 0;
		
		for (int i = 0; i < instances.size(); i++){
			String instance = instances.get(i);
			
			try{
				String instance_path = input_pattern.replace("{subject}", instance);
				File input_file = new File(instance_path);
				BufferedReader reader = new BufferedReader(new FileReader(input_file));
				
				String line = reader.readLine();
				
				if (i == 0){
					//first run
					while (line != null){
						sums.add(Double.valueOf(line));
						line = reader.readLine();
						}
				}else{
					int j = 0;
					while (line != null && j < sums.size()){
						sums.set(j, sums.get(j)+ Double.valueOf(line));
						line = reader.readLine();
						j++;
						}
					}
				
				reader.close();
				success_count++;
				log("average_surface_data: added instance '" + instance + "'.");
				
			}catch (IOException e){
				fail_count++;
				log("average_surface_data: error adding instance '" + instance + "'.");
				}
			
			}
		
		try{
			BufferedWriter writer = new BufferedWriter(new FileWriter(output_file));
			DecimalFormat df = new DecimalFormat(number_format);
			for (int i = 0; i < sums.size(); i++){
				if (i > 0) writer.write("\n");
				writer.write(df.format(sums.get(i) / (double)success_count));
				}
			writer.close();
			log("average_surface_data: Wrote average data to '" + output_file.getAbsolutePath() + 
					"' (" + success_count + " successes " + ", " + fail_count + " failures).");
			
		}catch (IOException e){
			log("average_surface_data: Error writing average data to '" + output_file.getAbsolutePath() + "'.");
			}
		
		return fail_count == 0;
	}
	
	/*********************************************
	 * Calculates a vertex-wise standard deviation for a set of surface data instances (e.g., cortical thickness).
	 * Instances are specified by an instance directory and a "\n"-delimited text file listing instance 
	 * names. If no list file is specified, all instances are used. 
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>instance_dir			The instance directory
	 * <li>input_file_pattern	Pattern specifying a path to each input data file, relative to each instance 
	 * 							subdirectory. Use "{subject}" to specify insert the current instance name into 
	 * 							the path.
	 * <li>output_file			The file to contain the averaged vertex values
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>average_data_file		A file containing the average vertex-wise values (i.e., as produced by 
	 * 								<code>average_surface_data()</code>; no such file indicates that the averages
	 * 								are to be calculated by the function.
	 * <li>instance_list_file		A text file listing the instances to average ("\n"-delimited); no file
	 * 								indicates that all instances are to be used
	 * <li>log_file					File to place log comments and errors; no log_file specifies that comments
	 * 								are to be sent to the console
	 * <li>number_format			Number format, e.g., "##0.000".
	 * </ul>
	 * 
	 */
	public boolean stdev_surface_data(){
		
		// Get subdirectories
		String param = parameters.get("instance_dir");
		if (param == null){
			log("stdev_surface_data: Parameter 'instance_dir' is required.");
			return false;
			}
		
		File instance_dir = new File(param);
		
		if (!instance_dir.exists() || !instance_dir.isDirectory()){
			log("stdev_surface_data: Directory '" + instance_dir.getAbsolutePath() + "' does not " +
														"exist or is not a directory.");
			return false;
			}
		
		String input_pattern = parameters.get("input_file_pattern");
		if (input_pattern == null){
			log("stdev_surface_data: Parameter 'input_file_pattern' is required.");
			return false;
			}
		
		input_pattern = input_pattern.replace("/", File.separator);
		input_pattern = input_pattern.replace("\\", File.separator);
		input_pattern = param + File.separator + input_pattern;
		
		param = parameters.get("average_data_file");
		File average_data_file = null;
		
		if (param != null){
			average_data_file = new File(param);
			if (!average_data_file.exists() || !average_data_file.isFile()){
				log("stdev_surface_data: Average data file '" + average_data_file.getAbsolutePath() +
										"' is invalid.");
				return false;
				}
			}
		
		param = parameters.get("output_file");
		if (param == null){
			log("stdev_surface_data: Parameter 'output_file' is required.");
			return false;
			}
		
		File output_file = new File(param);
		
		try{
			if (!output_file.exists() && !output_file.createNewFile()){
				log("stdev_surface_data: File '" + output_file.getAbsolutePath() + "' could not be created.");
				return false;
				}
		}catch (IOException e){
			log("stdev_surface_data: File '" + output_file.getAbsolutePath() + "' could not be created.");
			return false;
			}
		
		String number_format = parameters.get("number_format");
		if (number_format == null){
			number_format = "#0.00000000";
			}
		
		ArrayList<String> instances = new ArrayList<String>();
		
		// Load subject list
		param = parameters.get("instance_list_file");
		if (param == null){
			log("stdev_surface_data: No instance file specified; using all instances.");
			String[] list = instance_dir.list(IoFunctions.getDirFilter());
			for (int i = 0; i < list.length; i++)
				instances.add(list[i]);
		}else{
			File list_file = new File(param);
			log("stdev_surface_data: Loading instances from '" + list_file.getAbsolutePath() + "'.");
			if (!list_file.exists()){
				log("stdev_surface_data: File '" + list_file.getAbsolutePath() + "' not found.");
				return false;
				}
			try{
				BufferedReader reader = new BufferedReader(new FileReader(list_file));
				String line = reader.readLine();
				while (line != null){
					instances.add(line);
					line = reader.readLine();
					}
				reader.close();
					
			}catch (IOException e){
				log("stdev_surface_data: Error reading file '" + list_file.getAbsolutePath() + "'.");
				return false;
				}
			}
		
		// Cross-reference; report errors
		
		
		// Get means
		ArrayList<Double> means = new ArrayList<Double>();
		if (average_data_file != null){
			try{
				BufferedReader reader = new BufferedReader(new FileReader(average_data_file));
				String line = reader.readLine();
				while (line != null){
					means.add(Double.valueOf(line));
					line = reader.readLine();
					}
			}catch (IOException e){
				log("stdev_surface_data: Error reading average data file '" + average_data_file.getAbsolutePath() + "'.");
				return false;
				}
			
		}else{
			int success_count = 0, fail_count = 0;
			ArrayList<Double> sums = new ArrayList<Double>();
		
			for (int i = 0; i < instances.size(); i++){
				String instance = instances.get(i);
				
				try{
					String instance_path = input_pattern.replace("{subject}", instance);
					File input_file = new File(instance_path);
					BufferedReader reader = new BufferedReader(new FileReader(input_file));
					
					String line = reader.readLine();
					
					if (i == 0){
						//first run
						while (line != null){
							sums.add(Double.valueOf(line));
							line = reader.readLine();
							}
					}else{
						int j = 0;
						while (line != null && j < sums.size()){
							sums.set(j, sums.get(j)+ Double.valueOf(line));
							line = reader.readLine();
							j++;
							}
						}
					
					reader.close();
					success_count++;
					log("average_surface_data: added instance '" + instance + "'.");
					
				}catch (IOException e){
					fail_count++;
					log("average_surface_data: error adding instance '" + instance + "'.");
					}
				
				}
			
			for (int i = 0; i < sums.size(); i++){
				means.add(sums.get(i) / (double)success_count);
				}
			
			log("stdev_surface_data: Calculated means (" + success_count + " successes, " + 
															fail_count + " failures).");
			}
		
		// Get sum-of-squares
		ArrayList<Double> sum_squares = new ArrayList<Double>();
		int success_count = 0, fail_count = 0;
		
		for (int i = 0; i < instances.size(); i++){
			String instance = instances.get(i);
			
			try{
				String instance_path = input_pattern.replace("{subject}", instance);
				File input_file = new File(instance_path);
				BufferedReader reader = new BufferedReader(new FileReader(input_file));
				
				String line = reader.readLine();
				int j = 0;
				
				if (i == 0){
					//first run
					while (line != null){
						sum_squares.add(Math.pow(Double.valueOf(line) - means.get(j), 2));
						line = reader.readLine();
						j++;
						}
				}else{
					while (line != null && j < sum_squares.size()){
						sum_squares.set(j, sum_squares.get(j) + Math.pow(Double.valueOf(line) - means.get(j), 2));
						line = reader.readLine();
						j++;
						}
					}
				
				reader.close();
				success_count++;
				log("stdev_surface_data: sum-of-squares, added instance '" + instance + "'.");
				
			}catch (IOException e){
				fail_count++;
				log("stdev_surface_data: sum-of-squares, error adding instance '" + instance + "'.");
				}
			}
		
		log("stdev_surface_data: Calculated sums of squares (" + success_count + " successes, " + 
				fail_count + " failures).");
		
		// Calculate standard deviation as sqrt(sum_sq / n)
		
		try{
			BufferedWriter writer = new BufferedWriter(new FileWriter(output_file));
			DecimalFormat df = new DecimalFormat(number_format);
			for (int i = 0; i < sum_squares.size(); i++){
				if (i > 0) writer.write("\n");
				writer.write(df.format(Math.sqrt(sum_squares.get(i) / (double)success_count)));
				}
			writer.close();
			log("stdev_surface_data: Wrote st dev data to '" + output_file.getAbsolutePath() + 
					"' (" + success_count + " successes " + ", " + fail_count + " failures).");
			
		}catch (IOException e){
			log("stdev_surface_data: Error writing st dev data to '" + output_file.getAbsolutePath() + "'.");
			}
		
		return fail_count == 0;
		
	}
	
	
	/***********************************************
	 * Calculates the surface area and volume, given a Minc-format surface and vertex-wise thickness
	 * values. 
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>surface_file			The surface file for which to calculate surface area
	 * <li>thickness_file		The vertex-wise thickness values
	 * <li>output_file			Path to the file where the result will be stored
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>append				Flag specifying whether to append to the output file; default is to overwrite it
	 * <li>write_header			Flag specifying whether to write a header line
	 * <li>subject_name			The name of the subject; if specified, it will be stored along with the
	 * 							result, in the format [subject_name]\t[result]
	 * <li>log_file				File to place log comments and errors; no log_file specifies that comments
	 * 							are to be sent to the console
	 * <li>number_format		Number format, e.g., "##0.000"
	 * </ul>
	 * 
	 */
	public boolean calculate_area_and_volume(){
		
		String number_format = parameters.get("number_format");
		if (number_format == null){
			number_format = "#0.00000000";
			}
		
		String subject_name = parameters.get("subject_name");
		boolean append = parameters.containsKey("append");
		boolean write_header = parameters.containsKey("write_header");
		
		String param = parameters.get("surface_file");
		if (param == null){
			log("calculate_area_and_volume: Parameter 'surface_file' is required.");
			return false;
			}
		
		File surface_file = new File(param);
		
		if (!surface_file.exists()){
			log("calculate_area_and_volume: Surface file '" + surface_file.getAbsolutePath() + "' does not exist.");
			return false;
			}
		
		param = parameters.get("thickness_file");
		if (param == null){
			log("calculate_area_and_volume: Parameter 'thickness_file' is required.");
			return false;
			}
		
		File thickness_file = new File(param);
		
		if (!thickness_file.exists()){
			log("calculate_area_and_volume: Thickness file '" + thickness_file.getAbsolutePath() + "' does not exist.");
			return false;
			}
		
		param = parameters.get("output_file");
		if (param == null){
			log("calculate_area_and_volume: Parameter 'output_file' is required.");
			return false;
			}
		
		File output_file = new File(param);
		boolean output_is_new = !output_file.exists();
		
		try{
			if (output_is_new){
				output_file.createNewFile();
				}
		}catch (IOException e){
			log("calculate_area_and_volume: Could not create output file '" + output_file.getAbsolutePath() + "'.");
			return false;
			}
		
		MincSurfaceLoader loader = new MincSurfaceLoader(surface_file);
		Mesh3DInt mesh_int = null;
		
		try{
			mesh_int = loader.loadSurface();
			if (mesh_int == null){
				log("calculate_area_and_volume: Could not load surface from '" + surface_file.getAbsolutePath() + "'.");
				return false;
				}
		}catch (IOException ex){
			log("calculate_area_and_volume: Could not load surface from '" + surface_file.getAbsolutePath() + "'.");
			return false;
			}
		
		MincSurfaceDataLoader t_loader = new MincSurfaceDataLoader(thickness_file);
		ArrayList<MguiNumber> thickness = t_loader.loadValues();
		
		Mesh3D mesh = mesh_int.getMesh();
		double[] values = MeshFunctions.getVolumeAndArea(mesh, thickness, null, 0);
		
		String subject = "";
		if (subject_name != null)
			subject = subject_name + "\t";
		
		try{
			BufferedWriter writer = new BufferedWriter(new FileWriter(output_file, append));
			if (append && !output_is_new) writer.write("\n");
			if (output_is_new && write_header){
				if (subject_name != null)
					writer.write("Subject\tArea\tVolume\n");
				else
					writer.write("Area\tVolume\n");
				}
			writer.write(subject + MguiDouble.getString(values[0], number_format) + "\t" + 
								   MguiDouble.getString(values[1], number_format));
			writer.close();
			
		}catch (IOException e){
			log("calculate_area_and_volume: Could not write results to '" + output_file.getAbsolutePath() + "'.");
			return false;
			}
		
		if (subject_name != null)
			subject = "for " + subject_name + " ";
		
		log("calculate_area_and_volume: Result " + subject + "written to '" + output_file.getAbsolutePath() + "'.");
		return true;
		
	}
	
	
	/***********************************************
	 * Creates a scatterplot for a specific vertex in a Civet surface. 
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>instances_dir		The directory containing Civet subject data
	 * <li>instance_x_data		The x-axis data, of the form "[subject-id]\t[value]" (e.g., age)
	 * <li>data_file_pattern	A pattern specifying a path to the subject data file (e.g., thickness)
	 * <li>output_file			Path to the file where the scatterplot image will be stored
	 * <li>vertex_index			Index of the vertex for which to obtain a scatterplot
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>parameter_file		File specifying the optional parameters, of the form "[parameter_name]\t
	 * 							[parameter_value]". These values will override any values specified in the
	 * 							command line.
	 * <li>instance_list_file	A text file listing the instances to average ("\n"-delimited); no file
	 * 							indicates that all instances are to be used
	 * <li>size_x				Image x size, in pixels
	 * <li>size_y				Image y size, in pixels
	 * <li>symbol				Symbol to plot; one of: filled_circle, circle, filled_square, square 
	 * <li>symbol_colour		Colour of the symbols
	 * <li>title				The title of the scatterplot; use {roi} to substitute the ROI name
	 * <li>regression_line		Flag indicating whether to draw a regression line
	 * <li>regression_colour	Colour of the regression line, of the form "[red],[green],[blue]", where
	 * 							values are integers from 0 to 255
	 * <li>legend				Flag indicating whether to draw a legend 
	 * <li>log_file				File to place log comments and errors; no log_file specifies that comments
	 * 							are to be sent to the console
	 * <li>number_format_x		Number format for the x-axis, e.g., "##0.000"
	 * <li>number_format_y		Number format for the y-axis, e.g., "##0.000"
	 * <li>label_x				Label for the x-axis
	 * <li>label_y				Label for the y-axis
	 * <li>range_x				The range of the x values (will use entire range if not specified)
	 * <li>range_y				The range of the y values (will use entire range if not specified)
	 * <li>min_x				The minimum x axis value (will plot around mean if not specified)
	 * <li>min_y				The minimum y axis value (will plot around mean if not specified)
	 * </ul>
	 * 
	 */
	public boolean vertex_scatterplot(){
		
		String param = parameters.get("parameter_file");
		if (param != null && !load_parameter_file(param))
				return false;
		
		param = parameters.get("instances_dir");
		if (param == null){
			log("vertex_scatterplot: Parameter 'subjects_dir' is required.");
			return false;
			}
		
		File instances_dir = new File(param);
		
		if (!instances_dir.exists() || !instances_dir.isDirectory()){
			log("vertex_scatterplot: Subject directory '" + instances_dir.getAbsolutePath() + "' is invalid.");
			return false;
			}
		
		param = parameters.get("instance_x_data");
		if (param == null){
			log("vertex_scatterplot: Parameter 'instance_x_data' is required.");
			return false;
			}
		
		File instance_x_data = new File(param);
		
		if (!instance_x_data.exists()){
			log("vertex_scatterplot: X data file '" + instance_x_data.getAbsolutePath() + "' does not exist.");
			return false;
			}
		
		param = parameters.get("vertex_index");
		if (param == null){
			log("calculate_area_and_volume: Parameter 'vertex_index' is required.");
			return false;
			}
		
		int vertex_index = Integer.valueOf(parameters.get("vertex_index"));
		
		param = parameters.get("output_file");
		if (param == null){
			log("vertex_scatterplot: Parameter 'output_file' is required.");
			return false;
			}
		
		File output_file = new File(param);
		
		String data_file_pattern = parameters.get("data_file_pattern");
		if (param == null){
			log("vertex_scatterplot: Parameter 'data_file_pattern' is required.");
			return false;
			}
		
		int size_x = 500;
		if (parameters.containsKey("size_x"))
			size_x = Integer.valueOf(parameters.get("size_x"));
		int size_y = 500;
		if (parameters.containsKey("size_y"))
			size_y = Integer.valueOf(parameters.get("size_y"));
		String title = "Scatterplot";
		if (parameters.containsKey("title"))
			title = parameters.get("title");
		Color symbol_colour = Color.blue;
		if (parameters.containsKey("symbol_colour")){
			String colour = parameters.get("symbol_colour");
			String[] parts = colour.split(",");
			if (parts.length != 3){
				log("vertex_scatterplot: invalid symbol colour [" + colour + "]");
				return false;
				}
			symbol_colour = new Color(Integer.valueOf(parts[0]),
									  Integer.valueOf(parts[1]),
									  Integer.valueOf(parts[2]));
			}
		boolean reg_line = parameters.containsKey("regression_line");
		Color reg_colour = symbol_colour;
		if (parameters.containsKey("reg_colour")){
			String colour = parameters.get("reg_colour");
			String[] parts = colour.split(",");
			if (parts.length != 3){
				log("vertex_scatterplot: invalid regression line colour [" + colour + "]");
				return false;
				}
			reg_colour = new Color(Integer.valueOf(parts[0]),
								   Integer.valueOf(parts[1]),
								   Integer.valueOf(parts[2]));
			}
		boolean legend = parameters.containsKey("legend");
		String label_x = "X";
		if (parameters.containsKey("label_x"))
			label_x = parameters.get("label_x");
		String label_y = "Y";
		if (parameters.containsKey("label_y"))
			label_y = parameters.get("label_y");
		double range_x = Double.NaN;
		if (parameters.containsKey("range_x"))
			range_x = Double.valueOf(parameters.get("range_x"));
		double range_y = Double.NaN;
		if (parameters.containsKey("range_y"))
			range_y = Double.valueOf(parameters.get("range_y"));
		double min_x = -1;
		if (parameters.containsKey("min_x"))
			min_x = Double.valueOf(parameters.get("min_x"));
		double min_y = -1;
		if (parameters.containsKey("min_y"))
			min_y = Double.valueOf(parameters.get("min_y"));
		
	
		//Instances
		ArrayList<String> instances = new ArrayList<String>();
		
		// Load subject list
		param = parameters.get("instance_list_file");
		if (param == null){
			log("vertex_scatterplot: No instance file specified; using all instances.");
			String[] list = instances_dir.list(IoFunctions.getDirFilter());
			for (int i = 0; i < list.length; i++)
				instances.add(list[i]);
		}else{
			File list_file = new File(param);
			log("vertex_scatterplot: Loading instances from '" + list_file.getAbsolutePath() + "'.");
			if (!list_file.exists()){
				log("vertex_scatterplot: File '" + list_file.getAbsolutePath() + "' not found.");
				return false;
				}
			try{
				BufferedReader reader = new BufferedReader(new FileReader(list_file));
				String line = reader.readLine();
				while (line != null){
					instances.add(line);
					line = reader.readLine();
					}
				reader.close();
					
			}catch (IOException e){
				log("vertex_scatterplot: Error reading instance list file '" + list_file.getAbsolutePath() + "'.");
				return false;
				}
			}
	
		HashMap<String, Double> x_data_map = new HashMap<String, Double>();
		
		//read X data
		try{
			BufferedReader reader = new BufferedReader(new FileReader(instance_x_data));
			String line = reader.readLine();
			
			while (line != null){
				String[] parts = line.split("\t", 2);
				x_data_map.put(parts[0], Double.valueOf(parts[1]));
				line = reader.readLine();
				}
			
			reader.close();
			
		}catch (Exception e){
			log("vertex_scatterplot: Error reading X data file '" + instance_x_data.getAbsolutePath() + "'.");
			return false;
			}
		
		int success_count = 0, fail_count = 0;
		File input_file = null;
		
		ArrayList<ArrayList<Double>> series = new ArrayList<ArrayList<Double>>();
		ArrayList<Double> x_data = new ArrayList<Double>();
		series.add(x_data);
		ArrayList<Double> y_data = new ArrayList<Double>();
		series.add(y_data);
		
		//read Y data
		for (int i = 0; i < instances.size(); i++){
			try{
				String instance = instances.get(i);
				if (x_data_map.containsKey(instance)){
					String path = data_file_pattern;
					while (path.contains("{subject}"))
						path = path.replace("{subject}", instance);
					
					input_file = new File(path);
					BufferedReader reader = new BufferedReader(new FileReader(path));
					for (int j = 0; j < vertex_index; j++)
						reader.readLine();
					double value = Double.valueOf(reader.readLine());
					y_data.add(value);
					x_data.add(x_data_map.get(instance));
					success_count++;
					log("vertex_scatterplot: Added Y data for subject '" + instance + "'.");
				}else{
					log("vertex_scatterplot: No Y data found for subject '" + instance + "' (skipping).");
					fail_count++;
					}
			
			}catch (Exception e){
				String file = "";
				if (input_file != null)
					file = " '" + input_file.getAbsolutePath() + "'";
				log("vertex_scatterplot: Error reading Y data file" + file + ".");
				fail_count++;
				}
			}
		
		ArrayList<String> header = new ArrayList<String>();
		header.add("1");
		ArrayList<Color> reg_colour_list = new ArrayList<Color>();
		reg_colour_list.add(reg_colour);
		ArrayList<Color> symbol_colour_list = new ArrayList<Color>();
		symbol_colour_list.add(symbol_colour);
		
		/*
		ChartFunctions.output_scatterplot_2(title, 
											output_file.getAbsolutePath(),
											header,
											series, 
											reg_line, 
											1, 
											size_x, size_y, 
											range_x, 
											range_y,
											min_x,
											min_y,
											-1,
											-1,
											label_x, label_y, 
											symbol_colour_list, 
											reg_colour_list,
											legend,
											false);
											*/
		
		log("vertex_scatterplot: Done. " + success_count + " successes. " + fail_count + " failures.");
		
		return fail_count == 0;
	}
	
	/***********************************************
	 * Creates scatterplots for a set of ROIs on a Civet surface, specified by a name file. 
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>instances_dir		The directory containing Civet subject data
	 * <li>instance_x_data		The x-axis data, of the form "[subject-id]\t[value]" (e.g., age)
	 * <li>y_data_file_pattern	A pattern specifying a path to the subject y data file (e.g., thickness); calculates
	 * 							means from vertex-wise data. Either this parameter or "y_mean_data_file" must be 
	 * 							provided, but not both.
	 * <li>y_mean_roi_file		A path to a ROI-wise mean subject y data file. Use {roi} to substitute the
	 * 							ROI name.
	 * 							Either this parameter or "y_data_file_pattern" must be provided, but not both.
	 * <li>output_file_pattern	Pattern specifying the path to the files where scatterplot images will be stored;
	 * 							use "{roi}" to substitute the ROI name
	 * <li>roi_file_pattern		Pattern specifying the path to vertex-wise ROI indexes
	 * <li>name_map_file		Name file associating ROI indexes with names
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>parameter_file			File specifying the optional parameters, of the form "[parameter_name]\t
	 * 								[parameter_value]". These values will override any values specified in the
	 * 								command line.
	 * <li>instance_list_file		A text file listing the instances to be used; no file
	 * 								indicates that all instances are to be used
	 * <li>multilevel_file			Optional tab-delimited file which will create multiple levels in the scatterplot; e.g., dividing by gender. 
	 * 								The <code>regression_colour</code> and <code>symbol_colour</code> parameters will be ignored if this parameter 
	 * 								is specified.
	 * <li>multilevel_colours_file	Specifies a colour file associating colours with each level, of the form: [level]\t[red]\t[green]\t[blue] 
	 * 								If not specified, colours will be assigned in the order: blue, red, green, orange, random.
	 * <li>switch_xy				Switch data for the x and y axes
	 * <li>size_x					Image x size, in pixels
	 * <li>size_y					Image y size, in pixels
	 * <li>symbol					Symbol to plot; one of: filled_circle, circle, filled_square, square 
	 * <li>symbol_colour			Colour of the symbols
	 * <li>title					The title of the scatterplot; use "{roi}" to substitute the ROI name
	 * <li>regression_line			Flag indicating whether to draw a regression line
	 * <li>regression_colour		Colour of the regression line, of the form "[red],[green],[blue]", where
	 * 								values are integers from 0 to 255
	 * <li>legend					Flag indicating whether to draw a legend 
	 * <li>log_file					File to place log comments and errors; no log_file specifies that comments
	 * 								are to be sent to the console
	 * <li>number_format_x			Number format for the x-axis, e.g., "##0.000"
	 * <li>number_format_y			Number format for the y-axis, e.g., "##0.000"
	 * <li>label_x					Label for the x-axis
	 * <li>label_y					Label for the y-axis
	 * <li>range_x					The range of the x values (will use entire range if not specified)
	 * <li>range_y					The range of the y values (will use entire range if not specified)
	 * <li>min_x					The minimum x axis value (will plot around mean if not specified)
	 * <li>min_y					The minimum y axis value (will plot around mean if not specified)
	 * </ul>
	 * 
	 */
	public boolean roi_scatterplots(){
		
		String param = parameters.get("parameter_file");
		if (param != null && !load_parameter_file(param))
				return false;
		
		boolean switch_xy = parameters.containsKey("switch_xy");
		
		param = parameters.get("multilevel_file");
		HashMap<String, String> multilevel_map = null;
		HashMap<String, Color> multilevel_colours = null;
		ArrayList<String> multilevel_names = null;
		if (param != null){
			File multilevel_file = new File(param);
			if (!multilevel_file.exists()){
				log("roi_scatterplots: Multilevel file '" + multilevel_file.getAbsolutePath() + "' is invalid.");
				return false;
				}
			log("roi_scatterplots: Multilevel plots.");
			
			try{
				TreeSet<String> names = new TreeSet<String>();
				multilevel_map = new HashMap<String, String>();
				multilevel_colours = new HashMap<String, Color>();
				
				param = parameters.get("multilevel_colours_file");
				BufferedReader reader;
				
				if (param != null){
					reader = new BufferedReader(new FileReader(param));
					String line = reader.readLine();
					
					while (line != null){
						StringTokenizer tokens = new StringTokenizer(line);
						multilevel_colours.put(tokens.nextToken(), new Color(Integer.valueOf(tokens.nextToken()),
																		     Integer.valueOf(tokens.nextToken()),
																		     Integer.valueOf(tokens.nextToken())));
						line = reader.readLine();
						}
					reader.close();
					}
				
				reader = new BufferedReader(new FileReader(multilevel_file));
				String line = reader.readLine();
				Color default_colour = Color.blue;
				
				while (line != null){
					StringTokenizer tokens = new StringTokenizer(line);
					String name = tokens.nextToken();
					String level = tokens.nextToken();
					names.add(level);
					multilevel_map.put(name, level);
					if (!multilevel_colours.containsKey(level)){
						//assign colours by default
						multilevel_colours.put(level, default_colour);
						if (default_colour.equals(Color.blue))
							default_colour = Color.red;
						else if (default_colour.equals(Color.blue))
							default_colour = Color.green;
						else if (default_colour.equals(Color.blue))
							default_colour = Color.orange;
						else
							default_colour = Colours.getRandom().getColor();
						}
					line = reader.readLine();
					}
				
				multilevel_names = new ArrayList<String>(names);
				reader.close();
				
			}catch (IOException e){
				log("roi_scatterplots: Could not read multilevel file '" + multilevel_file.getAbsolutePath() + "'.");
				return false;
				}
		}else{
			log("roi_scatterplots: Single level plots.");
			}
		
		param = parameters.get("instances_dir");
		if (param == null){
			log("roi_scatterplots: Parameter 'subjects_dir' is required.");
			return false;
			}
		
		File instances_dir = new File(param);
		
		if (!instances_dir.exists() || !instances_dir.isDirectory()){
			log("roi_scatterplots: Subject directory '" + instances_dir.getAbsolutePath() + "' is invalid.");
			return false;
			}
		
		param = parameters.get("instance_x_data");
		if (param == null){
			log("roi_scatterplots: Parameter 'instance_x_data' is required.");
			return false;
			}
		
		File instance_x_data = new File(param);
		
		if (!instance_x_data.exists()){
			log("roi_scatterplots: X data file '" + instance_x_data.getAbsolutePath() + "' does not exist.");
			return false;
			}
		
		String roi_file_pattern = parameters.get("roi_file_pattern");
		if (roi_file_pattern == null){
			log("roi_scatterplots: Parameter 'roi_file_pattern' is required.");
			return false;
			}
		
		param = parameters.get("name_map_file");
		if (param == null){
			log("roi_scatterplots: Parameter 'name_map_file' is required.");
			return false;
			}
		
		File name_map_file = new File(param);
		
		if (!name_map_file.exists()){
			log("roi_scatterplots: Name map file '" + name_map_file.getAbsolutePath() + "' does not exist.");
			return false;
			}
		
		String output_file_pattern = parameters.get("output_file_pattern");
		if (output_file_pattern == null){
			log("roi_scatterplots: Parameter 'output_file_pattern' is required.");
			return false;
			}
		
		String y_data_file_pattern = parameters.get("y_data_file_pattern");
		String y_mean_roi_file = null;
		if (y_data_file_pattern == null){
			y_mean_roi_file = parameters.get("y_mean_roi_file");
			if (y_mean_roi_file == null){
				log("roi_scatterplots: Parameter 'y_data_file_pattern' OR 'y_mean_roi_file' is required.");
				return false;
				}
			}
		
		int size_x = 500;
		if (parameters.containsKey("size_x"))
			size_x = Integer.valueOf(parameters.get("size_x"));
		int size_y = 500;
		if (parameters.containsKey("size_y"))
			size_y = Integer.valueOf(parameters.get("size_y"));
		String title = "Scatterplot";
		if (parameters.containsKey("title"))
			title = parameters.get("title");
		Color symbol_colour = Color.blue;
		if (parameters.containsKey("symbol_colour")){
			String colour = parameters.get("symbol_colour");
			String[] parts = colour.split(",");
			if (parts.length != 3){
				log("roi_scatterplots: invalid symbol colour [" + colour + "]");
				return false;
				}
			symbol_colour = new Color(Integer.valueOf(parts[0]),
									  Integer.valueOf(parts[1]),
									  Integer.valueOf(parts[2]));
			}
		boolean reg_line = parameters.containsKey("regression_line");
		Color reg_colour = symbol_colour;
		if (parameters.containsKey("reg_colour")){
			String colour = parameters.get("reg_colour");
			String[] parts = colour.split(",");
			if (parts.length != 3){
				log("roi_scatterplots: invalid regression line colour [" + colour + "]");
				return false;
				}
			reg_colour = new Color(Integer.valueOf(parts[0]),
								   Integer.valueOf(parts[1]),
								   Integer.valueOf(parts[2]));
			}
		boolean legend = parameters.containsKey("legend");
		String label_x = "X";
		if (parameters.containsKey("label_x"))
			label_x = parameters.get("label_x");
		String label_y = "Y";
		if (parameters.containsKey("label_y"))
			label_y = parameters.get("label_y");
		double range_x = Double.NaN;
		if (parameters.containsKey("range_x"))
			range_x = Double.valueOf(parameters.get("range_x"));
		double range_y = Double.NaN;
		if (parameters.containsKey("range_y"))
			range_y = Double.valueOf(parameters.get("range_y"));
		double min_x = Double.NaN;
		if (parameters.containsKey("min_x"))
			min_x = Double.valueOf(parameters.get("min_x"));
		double min_y = Double.NaN;
		if (parameters.containsKey("min_y"))
			min_y = Double.valueOf(parameters.get("min_y"));
		
	
		//Instances
		ArrayList<String> instances = new ArrayList<String>();
		
		// Load subject list
		param = parameters.get("instance_list_file");
		if (param == null){
			log("roi_scatterplots: No instance file specified; using all instances.");
			String[] list = instances_dir.list(IoFunctions.getDirFilter());
			for (int i = 0; i < list.length; i++)
				instances.add(list[i]);
		}else{
			File list_file = new File(param);
			log("roi_scatterplots: Loading instances from '" + list_file.getAbsolutePath() + "'.");
			if (!list_file.exists()){
				log("roi_scatterplots: File '" + list_file.getAbsolutePath() + "' not found.");
				return false;
				}
			try{
				BufferedReader reader = new BufferedReader(new FileReader(list_file));
				String line = reader.readLine();
				while (line != null){
					instances.add(line);
					line = reader.readLine();
					}
				reader.close();
					
			}catch (IOException e){
				log("roi_scatterplots: Error reading instance list file '" + list_file.getAbsolutePath() + "'.");
				return false;
				}
			}
	
		HashMap<String, Double> x_data_map = new HashMap<String, Double>();
		
		//read X data
		try{
			BufferedReader reader = new BufferedReader(new FileReader(instance_x_data));
			String line = reader.readLine();
			
			while (line != null){
				String[] parts = line.split("\t", 2);
				x_data_map.put(parts[0], Double.valueOf(parts[1]));
				line = reader.readLine();
				}
			
			reader.close();
			
		}catch (Exception e){
			log("roi_scatterplots: Error reading X data file '" + instance_x_data.getAbsolutePath() + "'.");
			return false;
			}
		
		//read name map
		NameMapLoader loader = new NameMapLoader(name_map_file);
		NameMap name_map = null;
		
		try{
			name_map = (NameMap)loader.loadObject();
			if (name_map == null){
				log("roi_scatterplots: Error reading name map file '" + name_map_file.getAbsolutePath() + "'.");
				return false;
				}
		}catch (IOException e){
			log("roi_scatterplots: Error reading name map file '" + name_map_file.getAbsolutePath() + "'.");
			return false;
			}
		
		int success_count = 0, fail_count = 0;
		
		// For each subject
			// Read data
			// Calculate ROI-wise means
			// Add to list of means
		
		ArrayList<String> rois = new ArrayList<String>(name_map.getNames());
		HashMap<String, ArrayList<ArrayList<Double>>> roi_data_series = new HashMap<String, ArrayList<ArrayList<Double>>>();
		
		for (int i = 0; i < rois.size(); i++){
			ArrayList<ArrayList<Double>> series = new ArrayList<ArrayList<Double>>();
			if (multilevel_names != null){
				for (int j = 0; j < multilevel_names.size(); j++){
					series.add(new ArrayList<Double>());
					series.add(new ArrayList<Double>());
					}
			}else{
				series.add(new ArrayList<Double>());
				series.add(new ArrayList<Double>());
				}
			roi_data_series.put(rois.get(i), series);
			}
		
		if (y_data_file_pattern != null){
			for (int i = 0; i < instances.size(); i++){
				String instance = instances.get(i);
				String instance_path = y_data_file_pattern;
				while (instance_path.contains("{subject}"))
					instance_path = instance_path.replace("{subject}", instance);
				String roi_path = roi_file_pattern;
				while (roi_path.contains("{subject}"))
					roi_path = roi_path.replace("{subject}", instance);
				
				File y_input_file = new File(instance_path);
				File roi_input_file = new File(roi_path);
				
				try{
					HashMap<String, Double> roi_y_means = new HashMap<String, Double>();
					HashMap<String, Integer> roi_y_counts = new HashMap<String, Integer>();
					
					BufferedReader y_reader = new BufferedReader(new FileReader(y_input_file));
					BufferedReader roi_reader = new BufferedReader(new FileReader(roi_input_file));
					
					String y_line = y_reader.readLine();
					String roi_line = roi_reader.readLine();
					
					while (y_line != null){
						String roi_in = name_map.get(Integer.valueOf(roi_line));
						if (roi_in != null && rois.contains(roi_in)){
							Double old = roi_y_means.get(roi_in);
							double d = Double.valueOf(y_line);
							
							if (old == null){
								roi_y_means.put(roi_in, d);
								roi_y_counts.put(roi_in, 1);
							}else{
								roi_y_means.put(roi_in, old + d);
								roi_y_counts.put(roi_in, roi_y_counts.get(roi_in) + 1);
								}
							}
							
						y_line = y_reader.readLine();
						roi_line = roi_reader.readLine();
						}
					
					y_reader.close();
					roi_reader.close();
					
					//add x and y data to series
					for (int r = 0; r < rois.size(); r++){
						String roi = rois.get(r);
						if (roi_y_means.containsKey(roi)){
							double mean = roi_y_means.get(roi) / (double)roi_y_counts.get(roi);
							
							ArrayList<ArrayList<Double>> series = roi_data_series.get(roi);
							
							if (multilevel_names != null){
								for (int j = 0; j < multilevel_names.size(); j++){
									if (multilevel_names.get(j).equals(multilevel_map.get(instance))){
										if (switch_xy){
											series.get((j * 2) + 1).add(x_data_map.get(instance));
											series.get(j * 2).add(mean);
										}else{
											series.get(j * 2).add(x_data_map.get(instance));
											series.get((j * 2) + 1).add(mean);
											}
										}
									}
							}else{
								if (switch_xy){
									series.get(1).add(mean);
									series.get(0).add(x_data_map.get(instance));
								}else{
									series.get(0).add(x_data_map.get(instance));
									series.get(1).add(mean);
									}
								}
							}
						}
					
					success_count++;
					log("roi_scatterplots: Read Y data for subject '" + instance + "'.");
				}catch (IOException e){
					log("roi_scatterplots: Error reading Y data for subject '" + instance + "'.");
					fail_count++;
					}
				}
			}else{
				
				for (int i = 0; i < rois.size(); i++){
					String roi = rois.get(i);
					String roi_path = y_mean_roi_file;
					while (roi_path.contains("{roi}"))
						roi_path = roi_path.replace("{roi}", roi);
					
					try{
						BufferedReader y_reader = new BufferedReader(new FileReader(roi_path));
						String y_line = y_reader.readLine();
						
						while (y_line != null){
							String instance = y_line.substring(0, y_line.indexOf("\t"));
							double y_value = Double.valueOf(y_line.substring(y_line.indexOf("\t") + 1));
							double x_value = x_data_map.get(instance);
							
							ArrayList<ArrayList<Double>> series = roi_data_series.get(roi);
							
							if (multilevel_names != null){
								for (int j = 0; j < multilevel_names.size(); j++){
									if (multilevel_names.get(j).equals(multilevel_map.get(instance))){
										if (switch_xy){
											series.get(j * 2).add(y_value);
											series.get((j * 2) + 1).add(x_value);
										}else{
											series.get(j * 2).add(x_value);
											series.get((j * 2) + 1).add(y_value);
											}
										}
									}
							}else{
								if (switch_xy){
									series.get(0).add(y_value);
									series.get(1).add(x_value);
								}else{
									series.get(0).add(x_value);
									series.get(1).add(y_value);
									}
								}
							
							y_line = y_reader.readLine();
							}
						
						log("roi_scatterplots: Read Y data for roi '" + roi + "'.");
						
					}catch (IOException ex){
						log("roi_scatterplots: Error reading Y data for roi '" + roi + "'.");
						fail_count++;
						}
					
					}
				
				}
		
		log("roi_scatterplots: Read Y data for all subjects (" + success_count + " successes, " + 
																 fail_count + " failures).");
		success_count = 0;
		fail_count = 0;	
		
		// For each ROI
			// Output scatterplot
		for (int r = 0; r < rois.size(); r++){
			
			String roi = rois.get(r);
			String roi_title = title;
			while (roi_title.contains("{roi}"))
				roi_title = roi_title.replace("{roi}", roi);
			
			String output = output_file_pattern;
			while (output.contains("{roi}"))
				output = output.replace("{roi}", roi);
			File output_file = new File(output);
			
			ArrayList<String> header = null;
			ArrayList<Color> reg_colour_list = new ArrayList<Color>();
			ArrayList<Color> symbol_colour_list = new ArrayList<Color>();
			
			if (multilevel_names != null){
				header = multilevel_names;
				for (int i = 0; i < multilevel_names.size(); i++){
					reg_colour_list.add(multilevel_colours.get(multilevel_names.get(i)));
					symbol_colour_list.add(multilevel_colours.get(multilevel_names.get(i)));
					}
			}else{
				header = new ArrayList<String>();
				header.add("1");
				reg_colour_list.add(reg_colour);
				symbol_colour_list.add(symbol_colour);
				}
			
			try{
				/*
				ChartFunctions.output_scatterplot_2(roi_title, 
													output_file.getAbsolutePath(),
													header,
													roi_data_series.get(roi), 
													reg_line, 
													1, 
													size_x, size_y, 
													range_x, 
													range_y,
													min_x,
													min_y,
													-1,
													-1,
													label_x, label_y, 
													symbol_colour_list, 
													reg_colour_list,
													legend,
													false);
													*/
				log("roi_scatterplots: Created plot for ROI '" + roi + "'.");
				success_count++;
			}catch (Exception e){
				log("roi_scatterplots: Error creating plot for ROI '" + roi + "'.");
				fail_count++;
				}
			}
		
		log("roi_scatterplots: Created scatterplots (" + success_count + " successes, " + 
				 										 fail_count + " failures).");
		
		return fail_count == 0;
		
	}
	
	/**********************************************************************************
	 * Converts integer values in a text file using a tab-delimited conversion file.
	 * 
	 * <p>Parameters
	 * 
	 * <ul>
	 * <li>{@code input_file} - One-column tab-delimited text file containing the values to be converted
	 * <li>{@code conversion_file} - Two-column tab-delimited text file containing values to convert; format is
	 * [original_value}\t[converted_value]
	 * <li>{@code output_file} - File in which to store the converted values
	 * <li>{@code unknown_value} [optional] - Value to use when no mapping exists; -1 by default
	 * </ul>
	 * 
	 * @return
	 */
	boolean convert_integer_values(){
		
		int unknown_value = -1;
		String _param = parameters.get("unknown_value");
		if (_param != null){
			try{
				unknown_value = Integer.valueOf(_param);
			}catch (NumberFormatException ex){
				log("convert_values: unknown_value is invalid:" + _param, LoggingType.Errors);
				return false;
				}
			}
		
		_param = parameters.get("input_file");
		if (_param == null){
			log("convert_values: input_file is a required parameter.", LoggingType.Errors);
			return false;
			}
		
		File input_file = new File(_param);
		if (!input_file.exists()){
			log("convert_values: Cannot find input file '" + input_file.getAbsolutePath() + "'.", LoggingType.Errors);
			return false;
			}
		
		_param = parameters.get("conversion_file");
		if (_param == null){
			log("convert_values: conversion_file is a required parameter.", LoggingType.Errors);
			return false;
			}
		
		File conversion_file = new File(_param);
		if (!conversion_file.exists()){
			log("convert_values: Cannot find conversion file '" + conversion_file.getAbsolutePath() + "'.", LoggingType.Errors);
			return false;
			}
		
		_param = parameters.get("output_file");
		if (_param == null){
			log("convert_values: output_file is a required parameter.", LoggingType.Errors);
			return false;
			}
		
		File output_file = new File(_param);
		if (output_file.exists() && !output_file.delete()){
			log("convert_values: Cannot overwrite output file '" + output_file.getAbsolutePath() + "'.", LoggingType.Errors);
			return false;
			}
		
		//Read conversion file
		HashMap<Integer,Integer> conversion_map = new HashMap<Integer,Integer>();
		try{
			int line_no = 0;
			BufferedReader reader = new BufferedReader(new FileReader(conversion_file));
			String line = reader.readLine();
			
			while (line != null){
				String[] parts = line.split("\t");
				
				if (parts.length != 2){
					log("convert_values: Bad line in conversion file at " + line_no + ".", LoggingType.Errors);
					return false;
					}
				
				conversion_map.put(Double.valueOf(parts[0]).intValue(), Double.valueOf(parts[1]).intValue());
				line = reader.readLine();
				}
			
			reader.close();
			
		}catch (IOException ex){
			log("convert_values: Cannot read conversion file '" + conversion_file.getAbsolutePath() + "'.", LoggingType.Errors);
			return false;
			}
		
		int line_no = 1;
		
		//Read input file & convert
		try{
			BufferedReader reader = new BufferedReader(new FileReader(input_file));
			BufferedWriter writer = new BufferedWriter(new FileWriter(output_file));
			String line = reader.readLine();
			
			while (line != null){
				int int_val = Double.valueOf(line).intValue();
				int new_val = unknown_value;
				if (conversion_map.containsKey(int_val))
					new_val = conversion_map.get(int_val);
				writer.write("" + new_val);
				
				line = reader.readLine();
				if (line != null)
					writer.write("\n");
				line_no++;
				}
			
			reader.close();
			writer.close();
			
			
		}catch (IOException ex){
			log("convert_values: Error reading/writing values.", LoggingType.Errors);
			return false;
		}catch (NumberFormatException ex){
			log("convert_values: Bad number format at line " + line_no, LoggingType.Errors);
			return false;
			}
		
		log("convert_values: Success. Result written to '" + output_file.getAbsolutePath() + "'.");
		return true;
	}
	
	
	
	
}
