package mgui.command;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

import mgui.numbers.MguiDouble;
import mgui.numbers.MguiNumber;
import mgui.stats.StatFunctions;

public class StatisticsFunctions extends CommandInstance {

	public static void main(String[] args) throws Exception{
		
		StatisticsFunctions instance = new StatisticsFunctions();
		instance.execute(args);
		
	}
	
	@Override
	public boolean run_command(String command){
		
		if (command.equals("holm-bonferroni-correction")){
			return holm_bonferroni_correction();
			}
		
		return false;
		
	}
	
	/*******************************************************
	 * Reads a tab-delimited text file containing p-values, creates a new column containing these values
	 * corrected with the Holm-Bonferroni method, and writes to a new file.
	 * 
	 * <p>Parameters [required]
	 * <ul>
	 * <li>input_file			The tab-delimited input file, with a header line
	 * <li>p_value_column		The header name for the column containing p-values
	 * <li>output_file			The output file containing an appended column with corrected p-values
	 * </ul>
	 * 
	 * <p>Parameters [optional]
	 * <ul>
	 * <li>log_file				File to place log comments and errors; no log_file specifies that comments
	 * 							are to be sent to the console
	 * <li>output_column		Name for corrected column; default is "hb_corrected"
	 * <li>number_format		Number format, e.g., "##0.000".
	 * </ul>
	 * 
	 */
	public boolean holm_bonferroni_correction(){
		
		String param = parameters.get("input_file");
		if (param == null){
			log("holm_bonferroni_correction: parameter 'input_file' is required.");
			return false;
			}
		
		File input_file = new File(param);
		
		if (!input_file.exists() || !input_file.isFile()){
			log("holm_bonferroni_correction: invalid input file '" + param + "'.");
			return false;
			}
		
		param = parameters.get("output_file");
		if (param == null){
			log("holm_bonferroni_correction: parameter 'output_file' is required.");
			return false;
			}
		
		File output_file = new File(param);
		
		if (output_file.exists() && !output_file.delete()){
			log("holm_bonferroni_correction: could not delete output file '" + param + "'.");
			return false;
			}
		
		try{
		if (!output_file.createNewFile()){
			log("holm_bonferroni_correction: could not create output file '" + param + "'.");
			return false;
			}
		}catch (IOException e){
			log("holm_bonferroni_correction: could not create output file '" + param + "'.");
			return false;
			}
		
		String p_value_column = parameters.get("p_value_column");
		if (p_value_column == null){
			log("holm_bonferroni_correction: parameter 'p_value_column' is required.");
			return false;
			}
		
		String output_column = parameters.get("output_column");
		if (output_column == null){
			output_column = "hb_corrected";
			}
		
		String number_format = parameters.get("number_format");
		if (number_format == null){
			number_format = "#0.0000000";
			}
		
		ArrayList<String> text_lines = new ArrayList<String>();
		ArrayList<MguiNumber> p_values = new ArrayList<MguiNumber>();
		String header = null;
		
		try{
			BufferedReader reader = new BufferedReader(new FileReader(input_file));
			header = reader.readLine();
			StringTokenizer tokens = new StringTokenizer(header);
			int p_column = -1;
			int i = 0;
			while (tokens.hasMoreTokens() && p_column < 0){
				if (tokens.nextToken().equals(p_value_column))
					p_column = i;
				i++;
				}
			
			if (p_column < 0){
				log("holm_bonferroni_correction: p-value column '" + p_value_column + "' not found.");
				return false;
				}

			//read values
			String line = reader.readLine();
			
			while (line != null){
				text_lines.add(line);
				tokens = new StringTokenizer(line);
				i = 0;
				while (tokens.hasMoreTokens()){
					String token = tokens.nextToken();
					if (i == p_column)
						p_values.add(new MguiDouble(token));
					i++;
					}
				
				line = reader.readLine();
				}
			
			reader.close();
			log("holm_bonferroni_correction: Finished reading input file.");
			
		}catch (Exception e){
			log("holm_bonferroni_correction: Problem reading '" + input_file.getAbsolutePath() + "'.");
			return false;
			}
		
		double[] corrected = StatFunctions.getHolmBonferroniCorrected(p_values);
		
		try{
			
			BufferedWriter writer = new BufferedWriter(new FileWriter(output_file));
			header = header + "\t" + output_column;
			writer.write(header);
			
			for (int i = 0; i < text_lines.size(); i++){
				writer.write("\n" + text_lines.get(i));
				writer.write("\t" + MguiDouble.getString(corrected[i], number_format));
				}
			
			writer.close();
			
		}catch (Exception e){
			log("holm_bonferroni_correction: Problem writing '" + output_file.getAbsolutePath() + "'.");
			return false;
			}
		
		log("holm_bonferroni_correction: Done. Output written to '" + output_file.getAbsolutePath() + "'." );
		return true;
	}

}
